@@ -1,13 +1,17 @@
 #include "termkey.h"
 #include "termkey-internal.h"
 
+#include <ctype.h>
 #include <errno.h>
 #include <poll.h>
 #include <unistd.h>
 #include <string.h>
+#include <strings.h>
 
 #include <stdio.h>
 
+#define strcaseeq(a,b) (strcasecmp(a,b) == 0)
+
 void termkey_check_version(int major, int minor)
 {
   if(major != TERMKEY_VERSION_MAJOR) {
@@ -141,6 +145,29 @@ static void print_key(TermKey *tk, TermKeyKey *key)
       fprintf(stderr, "Mouse ev=%d button=%d pos=(%d,%d)\n", ev, button, line, col);
     }
     break;
+  case TERMKEY_TYPE_POSITION:
+    {
+      int line, col;
+      termkey_interpret_position(tk, key, &line, &col);
+      fprintf(stderr, "Position report pos=(%d,%d)\n", line, col);
+    }
+    break;
+  case TERMKEY_TYPE_MODEREPORT:
+    {
+      int initial, mode, value;
+      termkey_interpret_modereport(tk, key, &initial, &mode, &value);
+      fprintf(stderr, "Mode report mode=%s %d val=%d\n", initial == '?' ? "DEC" : "ANSI", mode, value);
+    }
+    break;
+  case TERMKEY_TYPE_DCS:
+    fprintf(stderr, "Device Control String");
+    break;
+  case TERMKEY_TYPE_OSC:
+    fprintf(stderr, "Operating System Control");
+    break;
+  case TERMKEY_TYPE_UNKNOWN_CSI:
+    fprintf(stderr, "unknown CSI\n");
+    break;
   }
 
   int m = key->modifiers;
@@ -173,6 +200,70 @@ static const char *res2str(TermKeyResult res)
 }
 #endif
 
+/* Similar to snprintf(str, size, "%s", src) except it turns CamelCase into
+ * space separated values
+ */
+static int snprint_cameltospaces(char *str, size_t size, const char *src)
+{
+  int prev_lower = 0;
+  size_t l = 0;
+  while(*src && l < size - 1) {
+    if(isupper(*src) && prev_lower) {
+      if(str)
+        str[l++] = ' ';
+      if(l >= size - 1)
+        break;
+    }
+    prev_lower = islower(*src);
+    str[l++] = tolower(*src++);
+  }
+  str[l] = 0;
+  /* For consistency with snprintf, return the number of bytes that would have
+   * been written, excluding '\0' */
+  while(*src) {
+    if(isupper(*src) && prev_lower) {
+      l++;
+    }
+    prev_lower = islower(*src);
+    src++; l++;
+  }
+  return l;
+}
+
+/* Similar to strcmp(str, strcamel, n) except that:
+ *    it compares CamelCase in strcamel with space separated values in str;
+ *    it takes char**s and updates them
+ * n counts bytes of strcamel, not str
+ */
+static int strpncmp_camel(const char **strp, const char **strcamelp, size_t n)
+{
+  const char *str = *strp, *strcamel = *strcamelp;
+  int prev_lower = 0;
+
+  for( ; (*str || *strcamel) && n; n--) {
+    char b = tolower(*strcamel);
+    if(isupper(*strcamel) && prev_lower) {
+      if(*str != ' ')
+        break;
+      str++;
+      if(*str != b)
+        break;
+    }
+    else
+      if(*str != b)
+        break;
+
+    prev_lower = islower(*strcamel);
+
+    str++;
+    strcamel++;
+  }
+
+  *strp = str;
+  *strcamelp = strcamel;
+  return *str - *strcamel;
+}
+
 static TermKey *termkey_alloc(void)
 {
   TermKey *tk = malloc(sizeof(TermKey));
@@ -193,6 +284,9 @@ static TermKey *termkey_alloc(void)
 
   tk->restore_termios_valid = 0;
 
+  tk->ti_getstr_hook = NULL;
+  tk->ti_getstr_hook_data = NULL;
+
   tk->waittime = 50; /* msec */
 
   tk->is_closed = 0;
@@ -231,10 +325,9 @@ static int termkey_init(TermKey *tk, const char *term)
     if(termkey_register_keyname(tk, keynames[i].sym, keynames[i].name) == -1)
       goto abort_free_keynames;
 
-  register_c0(tk, TERMKEY_SYM_BACKSPACE, 0x08, NULL);
-  register_c0(tk, TERMKEY_SYM_TAB,       0x09, NULL);
-  register_c0(tk, TERMKEY_SYM_ENTER,     0x0d, NULL);
-  register_c0(tk, TERMKEY_SYM_ESCAPE,    0x1b, NULL);
+  register_c0(tk, TERMKEY_SYM_TAB,    0x09, NULL);
+  register_c0(tk, TERMKEY_SYM_ENTER,  0x0d, NULL);
+  register_c0(tk, TERMKEY_SYM_ESCAPE, 0x1b, NULL);
 
   struct TermKeyDriverNode *tail = NULL;
 
@@ -300,19 +393,14 @@ TermKey *termkey_new(int fd, int flags)
   tk->fd = fd;
 
   if(!(flags & (TERMKEY_FLAG_RAW|TERMKEY_FLAG_UTF8))) {
-    int locale_is_utf8 = 0;
     char *e;
 
-    if((e = getenv("LANG")) && strstr(e, "UTF-8"))
-      locale_is_utf8 = 1;
-
-    if(!locale_is_utf8 && (e = getenv("LC_MESSAGES")) && strstr(e, "UTF-8"))
-      locale_is_utf8 = 1;
-
-    if(!locale_is_utf8 && (e = getenv("LC_ALL")) && strstr(e, "UTF-8"))
-      locale_is_utf8 = 1;
-
-    if(locale_is_utf8)
+    /* Most OSes will set .UTF-8. Some will set .utf8. Try to be fairly
+     * generous in parsing these
+     */
+    if(((e = getenv("LANG")) || (e = getenv("LC_MESSAGES")) || (e = getenv("LC_ALL"))) &&
+       (e = strchr(e, '.')) && e++ &&
+       (strcaseeq(e, "UTF-8") || strcaseeq(e, "UTF8")))
       flags |= TERMKEY_FLAG_UTF8;
     else
       flags |= TERMKEY_FLAG_RAW;
@@ -325,7 +413,7 @@ TermKey *termkey_new(int fd, int flags)
   if(!termkey_init(tk, term))
     goto abort;
 
-  if(!termkey_start(tk))
+  if(!(flags & TERMKEY_FLAG_NOSTART) && !termkey_start(tk))
     goto abort;
 
   return tk;
@@ -350,9 +438,14 @@ TermKey *termkey_new_abstract(const char *term, int flags)
     return NULL;
   }
 
-  termkey_start(tk);
+  if(!(flags & TERMKEY_FLAG_NOSTART) && !termkey_start(tk))
+    goto abort;
 
   return tk;
+
+abort:
+  free(tk);
+  return NULL;
 }
 
 void termkey_free(TermKey *tk)
@@ -379,6 +472,12 @@ void termkey_destroy(TermKey *tk)
   termkey_free(tk);
 }
 
+void termkey_hook_terminfo_getstr(TermKey *tk, TermKey_Terminfo_Getstr_Hook *hookfn, void *data)
+{
+  tk->ti_getstr_hook = hookfn;
+  tk->ti_getstr_hook_data = data;
+}
+
 int termkey_start(TermKey *tk)
 {
   if(tk->is_started)
@@ -391,7 +490,11 @@ int termkey_start(TermKey *tk)
       tk->restore_termios_valid = 1;
 
       termios.c_iflag &= ~(IXON|INLCR|ICRNL);
-      termios.c_lflag &= ~(ICANON|ECHO);
+      termios.c_lflag &= ~(ICANON|ECHO
+#ifdef IEXTEN
+          | IEXTEN
+#endif
+      );
       termios.c_cc[VMIN] = 1;
       termios.c_cc[VTIME] = 0;
 
@@ -646,7 +749,13 @@ static TermKeyResult parse_utf8(const unsigned char *bytes, size_t len, long *cp
 
 static void emit_codepoint(TermKey *tk, long codepoint, TermKeyKey *key)
 {
-  if(codepoint < 0x20) {
+  if(codepoint == 0) {
+    // ASCII NUL = Ctrl-Space
+    key->type = TERMKEY_TYPE_KEYSYM;
+    key->code.sym = TERMKEY_SYM_SPACE;
+    key->modifiers = TERMKEY_KEYMOD_CTRL;
+  }
+  else if(codepoint < 0x20) {
     // C0 range
     key->code.codepoint = 0;
     key->modifiers = 0;
@@ -660,7 +769,7 @@ static void emit_codepoint(TermKey *tk, long codepoint, TermKeyKey *key)
       key->type = TERMKEY_TYPE_UNICODE;
       /* Generically modified Unicode ought not report the SHIFT state, or else
        * we get into complicationg trying to report Shift-; vs : and so on...
-       * In order to be able to represent Ctrl-Shift-A as CTRL modified 
+       * In order to be able to represent Ctrl-Shift-A as CTRL modified
        * unicode A, we need to call Ctrl-A simply 'a', lowercase
        */
       if(codepoint+0x40 >= 'A' && codepoint+0x40 <= 'Z')
@@ -710,15 +819,15 @@ void termkey_canonicalise(TermKey *tk, TermKeyKey *key)
   int flags = tk->canonflags;
 
   if(flags & TERMKEY_CANON_SPACESYMBOL) {
-    if(key->type == TERMKEY_TYPE_UNICODE && key->code.number == 0x20) {
+    if(key->type == TERMKEY_TYPE_UNICODE && key->code.codepoint == 0x20) {
       key->type     = TERMKEY_TYPE_KEYSYM;
       key->code.sym = TERMKEY_SYM_SPACE;
     }
   }
   else {
     if(key->type == TERMKEY_TYPE_KEYSYM && key->code.sym == TERMKEY_SYM_SPACE) {
-      key->type        = TERMKEY_TYPE_UNICODE;
-      key->code.number = 0x20;
+      key->type           = TERMKEY_TYPE_UNICODE;
+      key->code.codepoint = 0x20;
       fill_utf8(key);
     }
   }
@@ -1096,7 +1205,7 @@ const char *termkey_get_keyname(TermKey *tk, TermKeySym sym)
   return "UNKNOWN";
 }
 
-char *termkey_lookup_keyname(TermKey *tk, const char *str, TermKeySym *sym)
+static const char *termkey_lookup_keyname_format(TermKey *tk, const char *str, TermKeySym *sym, TermKeyFormat format)
 {
   /* We store an array, so we can't do better than a linear search. Doesn't
    * matter because user won't be calling this too often */
@@ -1106,17 +1215,29 @@ char *termkey_lookup_keyname(TermKey *tk, const char *str, TermKeySym *sym)
     if(!thiskey)
       continue;
     size_t len = strlen(thiskey);
-    if(strncmp(str, thiskey, len) == 0)
-      return (char *)str + len;
+    if(format & TERMKEY_FORMAT_LOWERSPACE) {
+      const char *thisstr = str;
+      if(strpncmp_camel(&thisstr, &thiskey, len) == 0)
+          return thisstr;
+    }
+    else {
+      if(strncmp(str, thiskey, len) == 0)
+        return (char *)str + len;
+    }
   }
 
   return NULL;
 }
 
+const char *termkey_lookup_keyname(TermKey *tk, const char *str, TermKeySym *sym)
+{
+  return termkey_lookup_keyname_format(tk, str, sym, 0);
+}
+
 TermKeySym termkey_keyname2sym(TermKey *tk, const char *keyname)
 {
   TermKeySym sym;
-  char *endp = termkey_lookup_keyname(tk, keyname, &sym);
+  const char *endp = termkey_lookup_keyname(tk, keyname, &sym);
   if(!endp || endp[0])
     return TERMKEY_SYM_UNKNOWN;
   return sym;
@@ -1161,6 +1282,10 @@ modnames[] = {
   { "Shift", "Alt",  "Ctrl" }, // LONGMOD
   { "S",     "M",    "C" },    // ALTISMETA
   { "Shift", "Meta", "Ctrl" }, // ALTISMETA+LONGMOD
+  { "s",     "a",    "c" },    // LOWERMOD
+  { "shift", "alt",  "ctrl" }, // LOWERMOD+LONGMOD
+  { "s",     "m",    "c" },    // LOWERMOD+ALTISMETA
+  { "shift", "meta", "ctrl" }, // LOWERMOD+ALTISMETA+LONGMOD
 };
 
 size_t termkey_strfkey(TermKey *tk, char *buffer, size_t len, TermKeyKey *key, TermKeyFormat format)
@@ -1169,12 +1294,15 @@ size_t termkey_strfkey(TermKey *tk, char *buffer, size_t len, TermKeyKey *key, T
   size_t l = 0;
 
   struct modnames *mods = &modnames[!!(format & TERMKEY_FORMAT_LONGMOD) +
-                                    !!(format & TERMKEY_FORMAT_ALTISMETA) * 2];
+                                    !!(format & TERMKEY_FORMAT_ALTISMETA) * 2 +
+                                    !!(format & TERMKEY_FORMAT_LOWERMOD) * 4];
 
   int wrapbracket = (format & TERMKEY_FORMAT_WRAPBRACKET) &&
                     (key->type != TERMKEY_TYPE_UNICODE || key->modifiers != 0);
 
-  if(format & TERMKEY_FORMAT_CARETCTRL && 
+  char sep = (format & TERMKEY_FORMAT_SPACEMOD) ? ' ' : '-';
+
+  if(format & TERMKEY_FORMAT_CARETCTRL &&
      key->type == TERMKEY_TYPE_UNICODE &&
      key->modifiers == TERMKEY_KEYMOD_CTRL) {
     long codepoint = key->code.codepoint;
@@ -1202,19 +1330,19 @@ size_t termkey_strfkey(TermKey *tk, char *buffer, size_t len, TermKeyKey *key, T
   }
 
   if(key->modifiers & TERMKEY_KEYMOD_ALT) {
-    l = snprintf(buffer + pos, len - pos, "%s-", mods->alt);
+    l = snprintf(buffer + pos, len - pos, "%s%c", mods->alt, sep);
     if(l <= 0) return pos;
     pos += l;
   }
 
   if(key->modifiers & TERMKEY_KEYMOD_CTRL) {
-    l = snprintf(buffer + pos, len - pos, "%s-", mods->ctrl);
+    l = snprintf(buffer + pos, len - pos, "%s%c", mods->ctrl, sep);
     if(l <= 0) return pos;
     pos += l;
   }
 
   if(key->modifiers & TERMKEY_KEYMOD_SHIFT) {
-    l = snprintf(buffer + pos, len - pos, "%s-", mods->shift);
+    l = snprintf(buffer + pos, len - pos, "%s%c", mods->shift, sep);
     if(l <= 0) return pos;
     pos += l;
   }
@@ -1226,10 +1354,17 @@ size_t termkey_strfkey(TermKey *tk, char *buffer, size_t len, TermKeyKey *key, T
     l = snprintf(buffer + pos, len - pos, "%s", key->utf8);
     break;
   case TERMKEY_TYPE_KEYSYM:
-    l = snprintf(buffer + pos, len - pos, "%s", termkey_get_keyname(tk, key->code.sym));
+    {
+      const char *name = termkey_get_keyname(tk, key->code.sym);
+      if(format & TERMKEY_FORMAT_LOWERSPACE)
+        l = snprint_cameltospaces(buffer + pos, len - pos, name);
+      else
+        l = snprintf(buffer + pos, len - pos, "%s", name);
+    }
     break;
   case TERMKEY_TYPE_FUNCTION:
-    l = snprintf(buffer + pos, len - pos, "F%d", key->code.number);
+    l = snprintf(buffer + pos, len - pos, "%c%d",
+        (format & TERMKEY_FORMAT_LOWERSPACE ? 'f' : 'F'), key->code.number);
     break;
   case TERMKEY_TYPE_MOUSE:
     {
@@ -1238,7 +1373,7 @@ size_t termkey_strfkey(TermKey *tk, char *buffer, size_t len, TermKeyKey *key, T
       int line, col;
       termkey_interpret_mouse(tk, key, &ev, &button, &line, &col);
 
-      static char *evnames[] = { "Unknown", "Press", "Drag", "Release" };
+      static const char *evnames[] = { "Unknown", "Press", "Drag", "Release" };
 
       l = snprintf(buffer + pos, len - pos, "Mouse%s(%d)",
           evnames[ev], button);
@@ -1263,6 +1398,12 @@ size_t termkey_strfkey(TermKey *tk, char *buffer, size_t len, TermKeyKey *key, T
       else
         l = snprintf(buffer + pos, len - pos, "Mode(%d=%d)", mode, value);
     }
+  case TERMKEY_TYPE_DCS:
+    l = snprintf(buffer + pos, len - pos, "DCS");
+    break;
+  case TERMKEY_TYPE_OSC:
+    l = snprintf(buffer + pos, len - pos, "OSC");
+    break;
   case TERMKEY_TYPE_UNKNOWN_CSI:
     l = snprintf(buffer + pos, len - pos, "CSI %c", key->code.number & 0xff);
     break;
@@ -1280,10 +1421,11 @@ size_t termkey_strfkey(TermKey *tk, char *buffer, size_t len, TermKeyKey *key, T
   return pos;
 }
 
-char *termkey_strpkey(TermKey *tk, const char *str, TermKeyKey *key, TermKeyFormat format)
+const char *termkey_strpkey(TermKey *tk, const char *str, TermKeyKey *key, TermKeyFormat format)
 {
   struct modnames *mods = &modnames[!!(format & TERMKEY_FORMAT_LONGMOD) +
-                                    !!(format & TERMKEY_FORMAT_ALTISMETA) * 2];
+                                    !!(format & TERMKEY_FORMAT_ALTISMETA) * 2 +
+                                    !!(format & TERMKEY_FORMAT_LOWERMOD) * 4];
 
   key->modifiers = 0;
 
@@ -1303,10 +1445,10 @@ char *termkey_strpkey(TermKey *tk, const char *str, TermKeyKey *key, TermKeyForm
     return (char *)str;
   }
 
-  const char *hyphen;
+  const char *sep_at;
 
-  while((hyphen = strchr(str, '-'))) {
-    size_t n = hyphen - str;
+  while((sep_at = strchr(str, (format & TERMKEY_FORMAT_SPACEMOD) ? ' ' : '-'))) {
+    size_t n = sep_at - str;
 
     if(n == strlen(mods->alt) && strncmp(mods->alt, str, n) == 0)
       key->modifiers |= TERMKEY_KEYMOD_ALT;
@@ -1318,14 +1460,14 @@ char *termkey_strpkey(TermKey *tk, const char *str, TermKeyKey *key, TermKeyForm
     else
       break;
 
-    str = hyphen + 1;
+    str = sep_at + 1;
   }
 
   size_t nbytes;
   ssize_t snbytes;
-  char *endstr;
+  const char *endstr;
 
-  if((endstr = termkey_lookup_keyname(tk, str, &key->code.sym))) {
+  if((endstr = termkey_lookup_keyname_format(tk, str, &key->code.sym, format))) {
     key->type = TERMKEY_TYPE_KEYSYM;
     str = endstr;
   }
@@ -1334,7 +1476,7 @@ char *termkey_strpkey(TermKey *tk, const char *str, TermKeyKey *key, TermKeyForm
     str += snbytes;
   }
   // Unicode must be last
-  else if(parse_utf8((unsigned char *)str, strlen(str), &key->code.codepoint, &nbytes) == TERMKEY_RES_KEY) {
+  else if(parse_utf8((unsigned const char *)str, strlen(str), &key->code.codepoint, &nbytes) == TERMKEY_RES_KEY) {
     key->type = TERMKEY_TYPE_UNICODE;
     fill_utf8(key);
     str += nbytes;
@@ -1390,6 +1532,9 @@ int termkey_keycmp(TermKey *tk, const TermKeyKey *key1p, const TermKeyKey *key2p
         return col1 - col2;
       }
       break;
+    case TERMKEY_TYPE_DCS:
+    case TERMKEY_TYPE_OSC:
+      return key1p - key2p;
     case TERMKEY_TYPE_MODEREPORT:
       {
         int initial1, initial2, mode1, mode2, value1, value2;
